from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import Group
from django.contrib.auth.admin import GroupAdmin as DjGroupAdmin
from django.contrib.auth.admin import UserAdmin as DjUserAdmin
from django.core.exceptions import ImproperlyConfigured

from table_permissions import app_settings
from table_permissions.widgets import TabularPermissionsWidget


User = get_user_model()


class UserTabularPermissionsMixin:

    def formfield_for_manytomany(self, db_field, request=None, **kwargs):
        field = super().formfield_for_manytomany(db_field, request, **kwargs)
        if db_field.name == "user_permissions":
            field.widget = TabularPermissionsWidget(
                db_field.verbose_name, db_field.name in self.filter_vertical
            )
            field.help_text = ""
        return field


class GroupTabularPermissionsMixin:
    def formfield_for_manytomany(self, db_field, request=None, **kwargs):
        field = super().formfield_for_manytomany(db_field, request, **kwargs)
        if db_field.name == "permissions":
            field.widget = TabularPermissionsWidget(
                db_field.verbose_name,
                db_field.name in self.filter_vertical,
                "permissions",
            )
            field.help_text = ""
        return field


try:
    UserAdminModel = admin.site._registry[User].__class__
except:  # noqa: E722
    UserAdminModel = DjUserAdmin

try:
    GroupAdminModel = admin.site._registry[Group].__class__
except:  # noqa: E722
    GroupAdminModel = DjGroupAdmin


class TabularPermissionsUserAdmin(UserTabularPermissionsMixin, UserAdminModel):
    pass


class TabularPermissionsGroupAdmin(GroupTabularPermissionsMixin, GroupAdminModel):
    pass


if app_settings.AUTO_IMPLEMENT:
    try:
        admin.site.unregister(User)
        admin.site.register(User, TabularPermissionsUserAdmin)
        admin.site.unregister(Group)
        admin.site.register(Group, TabularPermissionsGroupAdmin)

    except:  # noqa: E722
        raise ImproperlyConfigured(
            "Please make sure that django.contrib.auth (Or the app containing your custom User model) "
            "comes before tabular_permissions in INSTALLED_APPS; Or set AUTO_IMPLEMENT to False in your settings."
        )



from collections import OrderedDict

from django import VERSION
from django.apps import apps
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.forms import SelectMultiple
from django.template.loader import get_template
from django.utils.encoding import force_str
from django.utils.safestring import mark_safe

from table_permissions.app_settings import APPS_CUSTOMIZATION_FUNC
from table_permissions.app_settings import CUSTOM_PERMISSIONS_CUSTOMIZATION_FUNC
from table_permissions.app_settings import EXCLUDE_APPS
from table_permissions.app_settings import EXCLUDE_FUNCTION
from table_permissions.app_settings import EXCLUDE_MODELS
from table_permissions.app_settings import JS_FILE
from table_permissions.app_settings import TEMPLATE
from table_permissions.app_settings import TRANSLATION_FUNC
from table_permissions.app_settings import USE_FOR_CONCRETE
from table_permissions.helpers import get_perm_name


def get_reminder_permissions_iterator(choices, reminder_perms):
    reminder_perms_list = reminder_perms.values()
    response = []
    for choice in choices:
        if choice[0] in reminder_perms_list:
            response.append(choice)
    return response


class TabularPermissionsWidget(FilteredSelectMultiple):
    class Media:
        js = (JS_FILE,)

    def __init__(
        self,
        verbose_name,
        is_stacked,
        input_name="user_permissions",
        attrs=None,
        choices=(),
    ):
        super().__init__(verbose_name, is_stacked, attrs, choices)
        self.managed_perms = []
        self.input_name = input_name  # in case of UserAdmin, it's 'user_permissions', GroupAdmin it's 'permissions'
        self.hide_original = True

    def render(self, name, value, attrs=None, renderer=None):
        choices = self.choices
        apps_available = OrderedDict()  # []  # main container to send to template
        user_permissions = Permission.objects.filter(id__in=value or []).values_list(
            "id", flat=True
        )
        all_perms = (
            Permission.objects.all()
            .values("id", "codename", "content_type_id")
            .order_by("codename")
        )
        excluded_perms = set()
        codename_id_map = {}
        for p in all_perms:
            codename_id_map["{}_{}".format(p["codename"], p["content_type_id"])] = p[
                "id"
            ]

        # reminder_perms used to detect if the tabular permissions covers all permissions,
        # if true, we don't need to make the default widget visible.
        reminder_perms = codename_id_map.copy()

        # a global flag to either show or hide the other permission column in the table
        custom_permissions_available = False

        for app in apps.get_app_configs():
            app_dict = {
                "verbose_name": force_str(app.verbose_name),
                "label": app.label,
                "models": OrderedDict(),
            }

            for model_name in app.models:
                model_custom_permissions = []
                model_custom_permissions_ids = []

                model = app.models[model_name]
                ct_id = ContentType.objects.get_for_model(
                    model, for_concrete_model=USE_FOR_CONCRETE
                ).pk

                view_perm_name = get_perm_name(model_name, "view")
                add_perm_name = get_perm_name(model_name, "add")
                change_perm_name = get_perm_name(model_name, "change")
                delete_perm_name = get_perm_name(model_name, "delete")

                view_perm_id = (
                    codename_id_map.get(f"{view_perm_name}_{ct_id}", False)
                    if "view" in model._meta.default_permissions
                    else False
                )
                add_perm_id = (
                    codename_id_map.get(f"{add_perm_name}_{ct_id}", False)
                    if "add" in model._meta.default_permissions
                    else False
                )
                change_perm_id = (
                    codename_id_map.get(f"{change_perm_name}_{ct_id}", False)
                    if "change" in model._meta.default_permissions
                    else False
                )
                delete_perm_id = (
                    codename_id_map.get(f"{delete_perm_name}_{ct_id}", False)
                    if "delete" in model._meta.default_permissions
                    else False
                )
                if model._meta.permissions:
                    custom_permissions_available = True
                    for codename, perm_name in model._meta.permissions:
                        c_perm_id = codename_id_map.get(f"{codename}_{ct_id}", False)
                        verbose_name = TRANSLATION_FUNC(codename, perm_name, ct_id)
                        model_custom_permissions.append(
                            (codename, verbose_name, c_perm_id)
                        )
                        model_custom_permissions_ids.append(c_perm_id)

                if (
                    view_perm_id
                    or add_perm_id
                    or change_perm_id
                    or delete_perm_id
                    or model_custom_permissions
                ):
                    excluded_perms.update(
                        [view_perm_id, add_perm_id, change_perm_id, delete_perm_id]
                    )
                    excluded_perms.update(*model_custom_permissions_ids)
                    reminder_perms.pop(f"{view_perm_name}_{ct_id}", False)
                    reminder_perms.pop(f"{add_perm_name}_{ct_id}", False)
                    reminder_perms.pop(f"{change_perm_name}_{ct_id}", False)
                    reminder_perms.pop(f"{delete_perm_name}_{ct_id}", False)
                    for custom_permission in model_custom_permissions:
                        reminder_perms.pop(f"{custom_permission[0]}_{ct_id}", False)

                    # Because the logic of exclusion should/would work on both the tabular_permissin widget
                    # and the normal widget
                    # ie bydefautlwe exclude the session, admin log permissions and we dont want that on either widgets
                    if (
                        app.label in EXCLUDE_APPS
                        or model_name in EXCLUDE_MODELS
                        or EXCLUDE_FUNCTION(model)
                    ):
                        continue

                    app_dict["models"][model_name] = {
                        "model_name": model_name,
                        "model": model,
                        "verbose_name_plural": force_str(
                            model._meta.verbose_name_plural
                        ),
                        "verbose_name": force_str(model._meta.verbose_name),
                        "view_perm_id": view_perm_id,
                        "view_perm_name": view_perm_name,
                        "add_perm_id": add_perm_id,
                        "add_perm_name": add_perm_name,
                        "change_perm_id": change_perm_id,
                        "change_perm_name": change_perm_name,
                        "delete_perm_id": delete_perm_id,
                        "delete_perm_name": delete_perm_name,
                        "custom_permissions": model_custom_permissions,
                    }

            if app.models:
                apps_available[app.label] = app_dict
        if VERSION >= (2, 1, 0) and custom_permissions_available:
            colspan = 7
        elif VERSION >= (2, 1, 0) or custom_permissions_available:
            colspan = 6
        else:
            colspan = 5

        apps_available = APPS_CUSTOMIZATION_FUNC(apps_available)
        request_context = {
            "apps_available": apps_available,
            "user_permissions": user_permissions,
            "codename_id_map": codename_id_map,
            "input_name": self.input_name,
            "custom_permissions_available": custom_permissions_available,
            "colspan": colspan,
            "django_supports_view_permissions": VERSION >= (2, 1, 0),
        }
        body = get_template(TEMPLATE).render(request_context).encode("utf-8")
        self.managed_perms = excluded_perms
        if reminder_perms:
            self.hide_original = False

        reminder_choices = get_reminder_permissions_iterator(choices, reminder_perms)

        # filter the left over permission
        reminder_choices = CUSTOM_PERMISSIONS_CUSTOMIZATION_FUNC(reminder_choices)
        if not reminder_choices:
            attrs["style"] = " display:none "
            # switching to "normal" SelectMultiple as FilteredSelectMultiple will render the widget even
            # if the style=display:none
            original_class = SelectMultiple(attrs, reminder_choices)
        else:
            original_class = FilteredSelectMultiple(
                self.verbose_name, self.is_stacked, attrs, reminder_choices
            )

        output = original_class.render(name, value, attrs, renderer)

        initial = mark_safe("".join(output))
        response = " <hr/>".join([force_str(body), force_str(initial)])
        return mark_safe(response)



from django.utils.translation import gettext_lazy as _


def get_perm_name(model_name, perm_name):
    return f"{perm_name}_{model_name}"


def dummy_permissions_exclude(model):
    """
    A dummy function that excludes nothing,
    :param model:
    :return:
    """
    return False


def custom_permissions_translator(codename, verbose_name, content_type_id):
    """
    A Hook for multi-lingual applications to translate custom permissions
    By default it will try to translated the permission verbose_name

    :param codename: permission codename
    :param verbose_name: permission verbose_name specified on model meta
    :param content_type_id: relevant permission content_type
    :return: a translated verbose_name
    """
    return _(verbose_name)


def apps_customization_func(apps):
    """
    This is a dummy method simulating a final apps adjustment
    :param apps: the OrderedDict
    :return: OrderedDict with teh custom order
    """

    return apps


def custom_permissions_customization_func(permissions):
    """
    What if you removed an app from the INSTALLED_APPS, its permissions will be displayed on the custom permission
    widget.
    This is a change to further customize what will appear.
    :param permissions:
    :return:
    """
    return permissions



from django.conf import settings
from django.utils.module_loading import import_string


TABULAR_PERMISSIONS_CONFIG = {
    "template": "tabular_permissions/admin/tabular_permissions.html",
    "js_file": "tabular_permissions/tabular_permissions.js",
    "exclude": {
        "override": False,
        "apps": [],
        "models": [],
        "function": "tabular_permissions.helpers.dummy_permissions_exclude",
    },
    "auto_implement": True,
    "use_for_concrete": False,
    "custom_permission_translation": "tabular_permissions.helpers.custom_permissions_translator",
    "apps_customization_func": "tabular_permissions.helpers.apps_customization_func",
    "custom_permissions_customization_func": "tabular_permissions.helpers.custom_permissions_customization_func",
}
user_conf = getattr(settings, "TABULAR_PERMISSIONS_CONFIG", False)

if user_conf:
    # we update the exclude dict first
    TABULAR_PERMISSIONS_CONFIG["exclude"].update(user_conf.get("exclude", {}))
    user_conf["exclude"] = TABULAR_PERMISSIONS_CONFIG["exclude"]
    # update the rest if the configuration
    TABULAR_PERMISSIONS_CONFIG.update(user_conf)

AUTO_IMPLEMENT = TABULAR_PERMISSIONS_CONFIG["auto_implement"]
TEMPLATE = TABULAR_PERMISSIONS_CONFIG["template"]
JS_FILE = TABULAR_PERMISSIONS_CONFIG["js_file"]

_base_exclude_apps = ["sessions", "contenttypes", "admin"]
user_exclude = TABULAR_PERMISSIONS_CONFIG["exclude"]
if not user_exclude.get("override", False):
    EXCLUDE_APPS = _base_exclude_apps + user_exclude.get("apps", [])
else:
    EXCLUDE_APPS = user_exclude.get("apps", [])
EXCLUDE_APPS = [x.lower() for x in EXCLUDE_APPS]

EXCLUDE_MODELS = user_exclude.get("models", [])
EXCLUDE_MODELS = [x.lower() for x in EXCLUDE_MODELS]

model_exclude_func = user_exclude.get("function")
EXCLUDE_FUNCTION = import_string(model_exclude_func)

USE_FOR_CONCRETE = TABULAR_PERMISSIONS_CONFIG["use_for_concrete"]
TRANSLATION_FUNC = import_string(
    TABULAR_PERMISSIONS_CONFIG["custom_permission_translation"]
)

APPS_CUSTOMIZATION_FUNC = import_string(
    TABULAR_PERMISSIONS_CONFIG["apps_customization_func"]
)
CUSTOM_PERMISSIONS_CUSTOMIZATION_FUNC = import_string(
    TABULAR_PERMISSIONS_CONFIG["custom_permissions_customization_func"]
)
